import numpy as np
from ..base import FunctionV2


class Rosenbrock(FunctionV2):
    def __init__(self, dim=20, **kwargs):
        super().__init__(**kwargs)

        self.dim = dim
        self.domain = np.tile([-2, 2], reps=(dim, 1))
        self.name = f"Rosenbrock (dim={dim})"
        self.glob_min = np.ones((1, dim))
        self.f_min = 0.0
    
    def func(self, x: np.ndarray) -> np.ndarray:
        return np.sum(
            100 * (x[..., 1:] - x[..., :-1] ** 2) ** 2 + (1 - x[..., :-1]) ** 2,
            axis=-1,
            keepdims=True
        )



from .base import Acquisition
from .acquisition import EI, CB, WU, WU_IDW, PI, ES, KG, BarycentreEI, GPBarycentreEI

__all__ = [
    'Acquisition',
    'EI',
    'CB',
    'WU',
    'WU_IDW',
    'PI',
    'ES',
    'KG',
    'BarycentreEI',
    'GPBarycentreEI'
]


import os
import pickle

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches

# from waggon.functions import FunctionV2 #creates circular import
from waggon.optim import Optimiser
from waggon.optim import SurrogateOptimiser


def load_results_for_plotting(func_dir, acqf_name, surr_name, base_dir='test_results', exp_transform=True):

    # print(func_dir, acqf_name, surr_name, base_dir)

    res_path = f'{base_dir}/{func_dir}/{acqf_name}/{surr_name}'

    # print(res_path)
    
    f = []
    for (_, _, filenames) in os.walk(res_path):
        f.extend(filenames)
        print(f)
        break
    
    # print(f)
    
    
    results = []
    
    # if len(f) == 0:
    #     return None, None
    
    for file in f:
        with open(f'{res_path}/{file}', "rb") as input_file:
            results.append(np.squeeze(pickle.load(input_file)))
    
    # if results == []:
    #     return np.ones(10), np.ones(10)
    if 'Bary' in acqf_name:
        return np.ones(10), 1e-1 * np.ones(10)
    # try:
    res = np.zeros((len(results), len(max(results, key=len))))
    # except:
    #     return np.ones(10), np.ones(10)
    
    for i in range(len(results)):
        res[i, :results[i].shape[0]] = results[i]
        
        if results[i].shape[0] < res.shape[0]:
            res[i, results[i].shape[0]:] = results[i][-1] * np.ones(res.shape[1] - results[i].shape[0])
    
    if exp_transform:
        res = np.exp(res)

    # print(res)
    
    return np.mean(res, axis=0), np.std(res, axis=0)


def display(source, ax=None, base_dir='test_results', y_label=False, x_label=False, title=None, **kwargs): # TODO: add option to plot single results

    results = {}
    # TODO: move args to saved results
    epsilon = kwargs['eps'] if 'eps' in kwargs else 1e-4
    max_iter = kwargs['max_iter'] if 'max_iter' in kwargs else 200

    if ax is None:
            fig, ax = plt.subplots(figsize=(10, 8))

    if type(source) == str: # if function directory is given; used when plotting all results
        res_path = f'{base_dir}/{source}'

        acqfs = []
        for (_, dirnames, _) in os.walk(res_path):
            acqfs.extend(dirnames)
            break
        
        j = 0

        for acqf in acqfs:
            surrs = []
            for (_, dirnames, _) in os.walk(f'{res_path}/{acqf}/'):
                surrs.extend(dirnames)
                break

            for surr in surrs:
                
                results[f'{surr} {acqf}'] = {'res': load_results_for_plotting(source, acqf, surr),
                                        'color': f'C{j}', 'label': f'{surr} ({acqf})'}
                
                j += 1 # TODO: fix colouring; works fine when the ran models are the same across all experiments

    elif type(source) in [Optimiser, SurrogateOptimiser]:
        results[f'{source.surr.name}'] = {'res': (np.array(source.errors), np.zeros_like(source.errors)),
                                          'color': 'C0', 'label': f'{source.surr.name} ({source.acqf.name})'}
        
        title = source.func.name if title is None else title
        epsilon = source.eps
        max_iter = source.max_iter
    
    # print(results)

    y_lims = []
    
    ax.hlines(epsilon, 0, max_iter, color='grey', linestyles='--', linewidth=5)

    for key in results.keys():
        
        ax.plot(np.arange(results[key]['res'][0].shape[0]), results[key]['res'][0], '-', linewidth=3, label=key, color=results[key]['color'])
        ax.fill_between(np.arange(results[key]['res'][0].shape[0]),
                        results[key]['res'][0] - results[key]['res'][1],
                        results[key]['res'][0] + results[key]['res'][1],
                        color=results[key]['color'], alpha=0.2)
        y_lims.append(results[key]['res'][0] - results[key]['res'][1])
    
    ax.set_title(source if title is None else title, fontsize=20, weight='bold')
    
    if y_label: ax.set_ylabel('Distance to optimum', fontsize=16, weight='bold')
    if x_label: ax.set_xlabel('Number of simulator calls', fontsize=16, weight='bold')
    
    try:
        y_lim = min(epsilon, np.min(y_lims))
    except ValueError:
        y_lim = -1.0
        pass
    
    ax.set_ylim(bottom=9e-1 * (y_lim if y_lim > 0 else epsilon))
    ax.tick_params(labelsize=16)
    ax.set_yscale('log')

    if type(source) in [Optimiser, SurrogateOptimiser]:
        handles = [mpatches.Patch(color=results[key]['color'], label=results[key]['label']) for key in results.keys()]
        fig.legend(handles=handles, bbox_to_anchor=(0.5, -0.1), loc='lower center', prop={'weight':'bold', 'size': 24})
        plt.show()
    
    return results


def plot_results(base_dir='test_results', figsize=(16, 8), **kwargs):

    funcs = []
    for (_, dirnames, _) in os.walk(base_dir):
        funcs.extend(dirnames)
        break
    
    nrows = int(np.round(np.sqrt(len(funcs))))
    nplots = len(funcs) if len(funcs) == nrows ** 2 else len(funcs) + 1
    ncols = int(np.round((nplots)/nrows))
    fig, ax = plt.subplots(ncols=ncols, nrows=nrows, figsize=figsize)
    
    for i, func_dir in enumerate(funcs):

        r, c = i // ncols, i % ncols
        results = display(source=func_dir, ax=ax if len(funcs) == 1 else (ax[c] if nrows == 1 else ax[r, c]), base_dir='test_results',
                              y_label=True if c == 0 else False,
                              x_label=True if ((r == nrows - 1) or ((r == nrows - 1) and (c == ncols - 1))) else False, kwargs=kwargs)

    handles = [mpatches.Patch(color=results[key]['color'], label=results[key]['label']) for key in results.keys()]

    if len(funcs) == nrows ** 2:
        fig.legend(handles=handles, bbox_to_anchor=(0.5, -0.26), loc='lower center', prop={'weight':'bold', 'size': 24})
    else:
        legend_ax = ax[-1] if nrows == 1 else ax[-1, -1]
        
        legend_ax.clear()
        legend_ax.set_axis_off()

        legend_ax.legend(handles=handles, loc='center', prop={'weight':'bold', 'size': 24})

    plt.tight_layout()
    if 'show' in kwargs:
        if kwargs['show']:
            plt.show()
    
    if not os.path.isdir('plots'):
        os.mkdir('plots')

    fig.savefig('plots/res.jpeg', format='jpeg', dpi=200, bbox_inches='tight')



import numpy as np
from ..base import FunctionV2


class Sphere(FunctionV2):
    def __init__(self, dim=1, **kwargs):
        super().__init__(**kwargs)
       
        self.dim = dim
        self.domain = np.tile([-10, 10], reps=(dim, 1))
        self.glob_min = np.zeros((1, dim))
        self.f_min = 0.0
        self.name = f"Sphere (dim={dim})"
    
    def func(self, x: np.ndarray) -> np.ndarray:
        return np.square(x).sum(axis=1, keepdims=True)



import numpy as np
from ..base import FunctionV2


class Himmelblau(FunctionV2):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.dim = 2
        self.domain = np.tile([-5, 5], reps=(2, 1))
        self.name = "Himmelblau"
        self.glob_min = np.array([
            [3.0, 2.0], 
            [-2.805118, 3.131312], 
            [-3.779310, -3.283186], 
            [3.584428, -1.848126]
        ])
        self.f_min = 0.0
    
    def func(self, x: np.ndarray) -> np.ndarray:
        return np.expand_dims(
            (x[..., 0] ** 2 + x[..., 1] - 11.0) ** 2 + (x[..., 0] + x[..., 1] ** 2 - 7.0) ** 2,
            axis=-1
        )



from abc import ABC
from abc import abstractmethod
from typing import Tuple

import numpy as np


class Function:
    def __init__(self, **kwargs):
        '''
        Base class for a synthetic black box.

        Parameters
        ----------
        dim : int, default = 1
            Dimensionality of the parameter space
        
        domain : np.array of shape (dim, 2), default = np.array([-10, 10])
            Contains bounds of the parameter space with the first column containing the smallest bound, and
            the second column containing the largest bound for the corresponding dimensions.
        
        name : str, default = 'parabola'
            Name of the function.
        
        glob_min : np.array of shape (dim,), default = np.zeros(dim)
            Point of global minimum of the function.
        
        f : Callable, default = lambda x: x**2
            The experiment function
        
        log_transform : bool, default = True
            Whether a log transformation is applied when calling the function or not.
        
        log_eps : float, default = 1e-8
            Small real value to prevent log(0) when log_transform is True.
        
        sigma : float, default = 1e-1
            Standard deviation of the output distribution.
        
        n_obs : int, default = 100
            Number of observations to sample for each point of the parameter space.
        '''
        super(Function, self).__init__()

        self.dim           = kwargs['dim'] if 'dim' in kwargs else 1 # TODO: add q-dimensional output?
        self.domain        = kwargs['domain'] if 'domain' in kwargs else np.array([[-10, 10]])
        self.name          = kwargs['name'] if 'name' in kwargs else 'parabola'
        self.glob_min      = kwargs['glob_min'] if 'glob_min' in kwargs else np.zeros(self.dim)
        self.f_min         = kwargs['f_min'] if 'f_min' in kwargs else 0.0
        self.f             = kwargs['f'] if 'f' in kwargs else lambda x: x**2
        self.log_transform = kwargs['log_transform'] if 'log_transform' in kwargs else True
        self.log_eps       = kwargs['log_eps'] if 'log_eps' in kwargs else 1e-8
        self.sigma         = kwargs['sigma'] if 'sigma' in kwargs else 1e-1
        self.n_obs         = kwargs['n_obs'] if 'n_obs' in kwargs else 1
    
    def __call__(self, x):
        '''
        Call of the black-box function

        Parameters
        ----------
        x : np.array of shape (n_samples, func.dim)
            Argument for which the function is called.
        
        Returns
        -------
        self.f(x) : np.array of shape (n_samples, func.dim)
            Black-box function values.
        '''
        if self.log_transform:
            return np.log(self.f(x) + self.log_eps)
        else:
            return self.f(x)

    def sample(self, x): # TODO: change to any distribution
        '''
        Sample from the black-box response distribution

        Parameters
        ----------
        x : np.array of shape (n_samples, func.dim)
            Argument for which the function is called.
        
        Returns
        -------
        X : np.array of shape (n_samples * func.n_obs, func.dim)
            Training input points. Correspond to input parameter x.

        y : np.array of shape (n_samples * func.n_obs, func.dim)
            Target values of the black-box function.
        '''

        y = np.random.normal(self.__call__(x[0, :].reshape(1, -1)), self.sigma, (self.n_obs, 1))
        X = x[0, :]*np.ones((self.n_obs, 1))
        
        for i in range(1, x.shape[0]):
            y_ = np.random.normal(self.__call__(x[i, :].reshape(1, -1)), self.sigma, (self.n_obs, 1))
            X_ = x[i, :]*np.ones((self.n_obs, 1))

            y = np.concatenate((y, y_))
            X = np.concatenate((X, X_))
        
        return X, y


class FunctionV2(ABC):
    def __init__(self, log_transform=False, log_eps=1e-7, n_obs=1, sampler='normal', **sampler_kwargs):
        self.log_transform = log_transform
        self.log_eps = log_eps
        self.n_obs = n_obs
        
        if sampler == "normal":
            self.sampler = sampler
            self.sigma = sampler_kwargs.pop('sigma', 1e-1)
        else:
            raise ValueError(
                f"{sampler} sampler is not supported now! :("
            )
        
        if len(sampler_kwargs) != 0:
            print(f"Not supported arguments: {sampler_kwargs.keys()}")
        
        if not hasattr(self, 'name'):
            self.name = self.__class__.__name__
        
    @abstractmethod
    def func(self, x: np.ndarray, *args, **kwargs) -> np.ndarray:
        pass

    def __call__(self, x: np.ndarray, *args, **kwargs) -> np.ndarray:
        if hasattr(self, "dim"):
            assert x.ndim == 2, f"Input array must be 2-dimensional, but it has shape {x.shape} with {x.ndim} dims"
            assert x.shape[-1] == self.dim, f"Mismatch between function ({self.dim}) and input ({x.shape[-1]}) dims"

        y = self.func(x, *args, **kwargs)

        if self.log_transform:
            return np.log(y + self.log_eps)
        return y
    
    def sample(self, x: np.ndarray, *args, **kwargs) -> Tuple[np.ndarray, np.ndarray]:
        y = self(x, *args, **kwargs)
        y_mu = np.repeat(y, repeats=self.n_obs, axis=-2)
        
        x_batch = np.repeat(x, repeats=self.n_obs, axis=-2)
        # Right now only normal sampler available
        y_batch = np.random.normal(loc=y_mu, scale=self.sigma, size=y_mu.shape)

        return x_batch, y_batch

    def __repr__(self) -> str:
        return self.name



import numpy as np
import scipy.linalg
from ..base import FunctionV2
from ..utils import fixed_numpy_seed


class SubmanifoldRosenbrock(FunctionV2):
    """
    Submanifold Rosenbrock problem function

    Parameters
    ----------
    dim: int, default 20
        Dimensionality of the function
    sub_dim: int, default 8
        Sub-dimensionality of the function
    seed: int | None, default None
        Seed for mapping matrix generation
    **kwargs
        Arguments passed to FunctionV2 base class
    
    Notes
    -----
    In case `seed=None`, NumPy automatically generates a seed 
    based on the system entropy source.
    """
    def __init__(self, dim=20, sub_dim=8, seed=None, **kwargs):
        super().__init__(**kwargs)

        self.dim = dim
        self.sub_dim = sub_dim
        self.domain = np.tile([-10, 10], reps=(dim, 1))
        self.name = f"SubmanifoldRosenbrock (dim={dim}, subdim={sub_dim})"

        with fixed_numpy_seed(seed):
            A = np.random.randn(dim, sub_dim)
        Q, _ = np.linalg.qr(A)
        self.Q = Q

        b = np.ones(sub_dim)

        x_min, _, _, _ = scipy.linalg.lstsq(Q.T, b)
        self.glob_min = np.expand_dims(x_min, 0)
        self.f_min = 0.0

    def func(self, x: np.ndarray) -> np.ndarray:
        y = np.matmul(x, self.Q)
        return np.sum(
            100.0 * (y[..., 1:] - (y[..., :-1]) ** 2.0) ** 2.0 + (1.0 - y[..., :-1]) ** 2,
            axis=-1,
            keepdims=True
        )


import os
import time
import pickle
import numpy as np
from tqdm import tqdm
from scipy.stats import qmc

from .utils import _get_olhs_num
from ..functions import Function


class Optimiser(object):
    def __init__(self, **kwargs):
        super(Optimiser, self).__init__()
        '''
        Black-box optimiser.

        Parameter
        ----------
        func : Callable, waggon.functions.Function
            Black-box function to be optimised.
        
        max_iter : int, default = 100
            Maximum number of optimisation loop iterations.
        
        eps : float, default = 1e-1
            Epsilon-solution criterion value.
        
        error_type : {'x', 'f'}, default = 'x'
            Optimisation error type - either in argument, 'x', or function, 'f'.
        
        eq_cons : dict, default = None
            Equality-type constraints for numerical optimisation.
        
        ineq_cons : dict, default = None
            Ineqaulity-type constraints for numerical optimisation.

        fix_candidates : bool, default = False if num_opt else True
            Whether the candidate points should be fixed or not.
        
        n_points : int, default = 1 if num_opt else 101**2
            Number of candidates points.
        
        olhs : bool, default = True
            Whether orthogonal Latin hypercube sampling (LHS) is used for choosing candidate points.
            If False, simple LHS is used.

        seed : int, default = None
            Controls the randomness of candidates sampled via (orthogonal) LHS.
        
        verbose : int, default = 1
            Controls verbosity when fitting and predicting. By default only a progress bar over the
            optimisation loop is displayed.
        
        save_results : bool, default = True
            Whether results are saved or not.
        '''
        self.func           = kwargs['func'] if 'func' in kwargs else Function()
        self.max_iter       = kwargs['max_iter'] if 'max_iter' in kwargs else 100
        self.eps            = kwargs['eps'] if 'eps' in kwargs else 1e-4
        self.error_type     = kwargs['error_type'] if 'error_type' in kwargs else 'x'
        self.n_points       = kwargs['n_points'] if 'n_points' in kwargs else 2 * self.func.dim
        self.olhs           = kwargs['olhs'] if 'olhs' in kwargs else False
        self.seed           = kwargs['seed'] if 'seed' in kwargs else None
        self.verbose        = kwargs['verbose'] if 'verbose' in kwargs else 1
        self.save_results   = kwargs['save_results'] if 'save_results' in kwargs else True
        self.plot_results   = kwargs['plot_results'] if 'plot_results' in kwargs else False

        if self.func.log_transform:
            transform = lambda x: np.exp(x)
        else:
            transform = lambda x: x
        
        if self.error_type == 'x':
            if self.func.glob_min is not None:
                self.error = lambda x: np.min(np.linalg.norm(self.func.glob_min - x, ord=2, axis=-1), axis=-1)
            else:
                raise ValueError
        elif self.error_type == 'f':
            if self.func.f_min is not None:
                self.error = lambda y: np.min(np.linalg.norm(self.func.f_min - transform(y), ord=2, axis=-1), axis=-1)
            elif self.func.glob_min is not None:
                self.func_f_min = self.func(self.func.glob_min)
                self.error = lambda y: np.min(np.linalg.norm(transform(self.func_f_min) - transform(y), ord=2, axis=-1), axis=-1)
            else:
                self.error = lambda y: np.max(transform(y))
        else:
            raise ValueError
            
    
    def create_candidates(self, N=None, strength=2):
        '''
        Creates candidate points among which the next best parameters will be selected.
        Also used for selecting the initial points of the optimisation process.

        Parameters
        ----------
        N : int, default = self.n_points
            Number of points to sample

        Returns
        -------
        candidates : np.array of shape (N, func.dim)
            Candidates points.
        '''
        
        if strength == 2:
            if N is None:
                N = _get_olhs_num((self.func.dim - 1)**2)[0]
        else:
            N = self.n_points if N is None else N

        lhs_       = qmc.LatinHypercube(d=self.func.domain.shape[0], scramble=True, strength=strength, seed=self.seed)
        candidates = lhs_.random(N)
        candidates = qmc.scale(candidates, self.func.domain[:, 0], self.func.domain[:, 1])
        return candidates
    
    def predict(self):
        pass
    
    def optimise(self, X=None, y=None):
        '''
        Runs the optimisation of the black-box function.

        Parameters
        ----------
        X : np.array of shape (n_samples * func.n_obs, func.dim), default = None
            Training input points. If None, an initial set of points and a corresponding training set
            will be created via self.create_candidates and func.sample functions respectively. Values will be created
            for both X and y.

        y : np.array of shape (n_samples * func.n_obs, func.dim), default = None
            Target values of the black-box function.
        '''

        self.errors = []
        
        if X is None:
            X = self.create_candidates(strength=1 if not self.olhs else 2)
            X, y = self.func.sample(X)
            self.res = np.array([[np.min(self.func(X))]])
            self.params = np.array([X[np.argmin(self.func(X)), :]])
        else:
            self.res = np.array([np.min(y)])
            self.params = np.array([X[np.argmin(y), :]])

        if self.verbose == 0:
            opt_loop = range(self.max_iter)
        else:
            error = self.error(y) if self.error_type == 'f' else self.error(X)
            opt_loop = tqdm(range(self.max_iter), desc=f"Optimisation error: {error:.4f}", leave=True, position=0)

        for _ in opt_loop:

            next_x = self.predict(X, y)

            X_, y_ = self.func.sample(next_x)
            X = np.concatenate((X, X_))
            y = np.concatenate((y, y_))

            if self.plot_results:
                self.plot_iteration_results(np.unique(X, axis=0), next_x[0])

            if y_ <= self.res[-1, :]:
                self.res = np.concatenate((self.res, y_.reshape(1, -1)))
                self.params = np.concatenate((self.params, next_x.reshape(1, -1)))
            else:
                self.res = np.concatenate((self.res, self.res[-1, :].reshape(1, -1)))
                self.params = np.concatenate((self.params, self.params[-1, :].reshape(1, -1)))

            error = self.error(y) if self.error_type == 'f' else self.error(X)
            self.errors.append(error)
            
            if self.verbose > 0:
                opt_loop.set_description(f"Optimisation error: {error:.4f}")
            
            if error <= self.eps:
                print('Experiment finished successfully!')
                break
        
        if error > self.eps:
            print('Experiment failed')
        
        if self.save_results:
            self._save()
    
    def _save(self, base_dir='test_results'):
        
        res = {'X': self.params,
               'y': self.res,
               'err': self.errors}

        if not os.path.isdir(base_dir):
            os.mkdir(base_dir)

        with open(f'{base_dir}/{time.strftime("%d_%m_%H_%M_%S")}.pkl', 'wb') as f:
            pickle.dump(res, f)
    
    def plot_iteration_results(self):
        pass



import numpy as np
from ..base import FunctionV2


class Ackley(FunctionV2):
    def __init__(self, dim=2, **kwargs):
        super().__init__(**kwargs)

        self.dim = dim
        self.domain = np.tile([-5, 5], reps=(dim, 1))
        self.glob_min = np.zeros((1, dim))
        self.f_min = 0.0
        self.name = f"Ackley (dim={dim})"
    
    def func(self, x: np.ndarray) -> np.ndarray:
        a, b, c, d = 20.0, 0.2, 2 * np.pi, self.dim
        x_sq_sum = np.square(x).sum(axis=-1, keepdims=True)
        x_cos_sum = np.cos(c * x).sum(axis=-1, keepdims=True)
        
        y = -a * np.exp(-b * np.sqrt(x_sq_sum / d))
        y -= np.exp(x_cos_sum / d)
        y += a + np.exp(1.0)
        return y



from .base import Surrogate

import GPy
import numpy as np

class GP(Surrogate):
    def __init__(self, **kwargs):
        super(GP, self).__init__()

        self.name     = 'GP'
        self.model    = kwargs['model'] if 'model' in kwargs else None
        self.kernel   = kwargs['kernel'] if 'kernel' in kwargs else None
        self.mean     = kwargs['mean'] if 'mean' in kwargs else None
        self.verbose  = kwargs['verbose'] if 'verbose' in kwargs else 1
    
    def fit(self, X, y):

        if self.model is None:
            
            if self.kernel is None:
                self.kernel = GPy.kern.Matern32(input_dim=X.shape[-1], lengthscale=1.0)

            self.model = GPy.models.GPRegression(X, y,
                                                 kernel = self.kernel,
                                                 mean_function = self.mean)
        
        self.model.set_XY(X=X, Y=y)
        
        self.model.optimize(messages=True if self.verbose > 1 else False)

    def predict(self, X):

        f, var = self.model.predict(X)
        std = np.sqrt(var)

        return f, std



from setuptools import setup, find_packages

with open("README.md", "r") as f:
    long_description = f.read()

setup(
    name="waggon",
    version="0.4.8",
    description="WAsserstein Global Gradient-free OptimisatioN (WAGGON) methods library.",
    packages=find_packages(),
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/hse-cs/waggon",
    author="Tigran Ramazyan",
    author_email="ramazyant@gmail.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.11",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.11",
    install_requires=[
        "GPy==1.13.2",
        "gpytorch==1.10.0", 
        "matplotlib==3.7.5", 
        "numpy==1.26.0",
        "scikit_learn==1.3.2", 
        "scipy==1.12.0", 
        "setuptools==72.1.0",
        "torch==2.2.2", 
        "torchbnn==1.2", 
        "torchensemble==0.1.9", 
        "tqdm==4.66.4", 
        # "jax==0.5.0", 
        # "protes==0.3.11"
    ],
    extras_require={
        "dev": [
            "pytest>=7.0", 
            "twine>=4.0.2",
        ],
    }
)



import numpy as np
import scipy.linalg

from .base import Function
from .utils import fixed_numpy_seed


class three_hump_camel(Function):
    '''
    Three Hump Camel function.
    '''
    def __init__(self, **kwargs):
        super(three_hump_camel, self).__init__(**kwargs)
        
        self.dim      = 2
        self.domain   = np.array([[-5, 5], [-5, 5]])
        self.name     = 'Three Hump Camel'
        self.glob_min = np.zeros(self.dim).reshape(1, -1)
        self.f_min    = 0.0
        self.f        = lambda x: 2 * x[:, 0]**2 - 1.05 * x[:, 0]**4 + x[:, 0]**6 / 6 + x[:, 0]*x[:, 1] + x[:, 1]**2


class rosenbrock(Function):
    '''
    d-dimensional Rosenbrock function.
    '''
    def __init__(self, dim=20, **kwargs):
        super(rosenbrock, self).__init__(**kwargs)

        self.dim      = dim
        self.domain   = np.array([self.dim*[-2, 2]]).reshape(self.dim, 2)
        self.name     = f'Rosenbrock ({self.dim} dim.)'
        self.glob_min = np.ones(self.dim).reshape(1, -1)
        self.f_min    = 0.0
        self.f        = lambda x: np.sum(np.array([100 * (x[:, i+1] - x[:, i] ** 2)**2 + (1 - x[:, i])**2 for i in range(self.dim - 1)]), axis=0)


class ackley(Function):
    '''
    Ackley function.
    '''
    def __init__(self, dim=2, **kwargs):
        super(ackley, self).__init__(**kwargs)

        self.dim      = dim
        self.domain   = np.array([self.dim*[-5, 5]]).reshape(self.dim, 2)
        self.name     = 'Ackley'
        self.glob_min = np.zeros(self.dim).reshape(1, -1)
        self.f_min    = 0.0
        self.f        = lambda x: -20 * np.exp(-0.2*np.sqrt((1./self.dim) * (np.sum(x**2, axis=1)))) -  np.exp((1./self.dim) * (np.sum(np.cos(2*np.pi*x), axis=1))) + np.e + 20


class levi(Function):
    '''
    Lévi function.
    '''
    def __init__(self, **kwargs):
        super(levi, self).__init__(**kwargs)

        self.dim      = 2
        self.domain   = np.array([self.dim*[-4, 6]]).reshape(self.dim, 2)
        self.name     = 'Lévi'
        self.glob_min = np.ones(self.dim).reshape(1, -1)
        self.f_min    = 0.0
        self.f        = lambda x: (np.sin(3*np.pi*x[:, 0]))**2 + ((x[:, 0] - 1)**2) * (1 + (np.sin(3*np.pi*x[:, 1]))**2) + ((x[:, 1] - 1)**2) * (1 + (np.sin(2*np.pi*x[:, 1]))**2)

    
    def sample(self, x):
        X, y = None, None
        
        for i in range(x.shape[0]):
            s1 = 0.04 - 0.03 * np.square(np.sin(3 * np.pi * x[i, 1]))
            s2 = 0.001 + 0.03 * np.square(np.sin(3 * np.pi * x[i, 1]))
            g1 = np.random.normal(self.__call__(x[i, :].reshape(1, -1))-0.05, s1, (self.n_obs//2, 1))
            g2 = np.random.normal(self.__call__(x[i, :].reshape(1, -1))+0.05, s2, (self.n_obs//2, 1))
            y_ = np.concatenate((g1, g2), axis=0)
            X_ = x[i, :]*np.ones((self.n_obs, 1))
        
            if i:
                X = np.concatenate((X, X_))
                y = np.concatenate((y, y_))
            else:
                X = X_
                y = y_
        
        return X, y


class himmelblau(Function):
    '''
    Himmelblau function.
    '''
    def __init__(self, **kwargs):
        super(himmelblau, self).__init__(**kwargs)

        self.dim      = 2
        self.domain   = np.array([self.dim*[-5, 5]]).reshape(self.dim, 2)
        self.name     = 'Himmelblau'
        self.glob_min = np.array([[3.0, 2.0], [-2.805118, 3.131312], [-3.779310, -3.283186], [3.584428, -1.848126]])
        self.f_min    = 0.0
        self.f        = lambda x: (x[:, 0]**2 + x[:, 1] - 11)**2 + (x[:, 0] + x[:, 1]**2 - 7)**2 


class holder(Function):
    '''
    Hölder function.
    '''
    def __init__(self, **kwargs):
        super(holder, self).__init__(**kwargs)

        self.dim      = 2
        self.domain   = np.array([self.dim*[-10, 10]]).reshape(self.dim, 2)
        self.name     = 'Holder'
        self.glob_min = np.array([[8.05502, 9.66459], [-8.05502, -9.66459], [-8.05502, 9.66459], [8.05502, -9.66459]])
        self.f_min    = 0.0
        self.f        = lambda x: -np.abs(np.sin(x[:, 0]) * np.cos(x[:, 1]) * np.exp(np.abs(1 - np.sqrt(x[:, 0]**2 + x[:, 1]**2)/np.pi))) + 19.2085


class submanifold_rosenbrock(Function):
    """
    Submanifold Rosenbrock problem function
    """
    def __init__(self, dim=20, sub_dim=8, seed=None, **kwargs):
        super().__init__(**kwargs)
        self.dim = dim
        self.sub_dim = sub_dim
        self.sigma = 1.0
        self.domain = np.array([self.dim*[-10, 10]]).reshape(self.dim, 2)
        
        self.name = f"Submanifold Rosenbrock (dim={dim}, subdim={sub_dim})"
        
        with fixed_numpy_seed(seed):
            A = np.random.randn(self.dim, self.sub_dim)
        Q, _ = np.linalg.qr(A)
        b = np.ones(sub_dim)

        x_min, _, _, _ = scipy.linalg.lstsq(Q.T, b)

        self.glob_min = np.expand_dims(x_min, 0)
        self.Q = Q
        self.f = lambda x: np.sum(
            100 * (x @ self.Q[:, 1:] - (x @ self.Q[:, :-1])**2)**2 + (1 - (x @ self.Q[:, :-1])) ** 2,
            axis=-1
        )


class tang(Function):
    '''
    d-dimensional Styblinsky-Tang function.
    '''
    def __init__(self, dim=20, **kwargs):
        super(tang, self).__init__(**kwargs)

        self.dim      = dim
        self.sigma    = 1.0
        self.domain   = np.array([self.dim*[-5, 5]]).reshape(self.dim, 2)
        self.name     = f'Styblinski-Tang ({self.dim} dim.)'
        self.glob_min = np.ones(self.dim).reshape(1, -1) * -2.903534

        self.f        = lambda x: np.sum(
            x ** 4.0 - 16.0 * x ** 2.0 + 5.0 * x, 
            axis=-1, 
        ) + 39.16617 * self.dim
        
        self.f_min    = 0.0
        # self.sigma    = lambda x: np.abs(x[:, 0] * np.sin(x[:, 1] - (-2.903534))) if 'sigma' in kwargs else lambda x: 1e-1
    
    # def __call__(self, x):
    #     if self.log_transform:
    #         if self.sigma(np.zeros((1, self.dim)))[0] == 1e-1:
    #             return np.log(self.f(x) + self.log_eps) 
    #         else:
    #             return np.log(self.f(x) + self.sigma(x) + self.log_eps) 
    #     else:
    #         if self.sigma(np.zeros((1, self.dim)))[0] == 1e-1:
    #             return self.f(x)
    #         else:
    #             return self.f(x) + self.sigma(x)

    # def sample(self, x):

    #     y = np.random.normal(self.__call__(x[0, :].reshape(1, -1)), self.sigma(x), (self.n_obs, 1))
    #     X = x[0, :]*np.ones((self.n_obs, 1))
        
    #     for i in range(1, x.shape[0]):
    #         y_ = np.random.normal(self.__call__(x[i, :].reshape(1, -1)), self.sigma(x), (self.n_obs, 1))
    #         X_ = x[i, :]*np.ones((self.n_obs, 1))

    #         y = np.concatenate((y, y_))
    #         X = np.concatenate((X, X_))
        
    #     return X, y


from .base import Optimiser
from .surrogate import SurrogateOptimiser
from .barycentre import BarycentreSurrogateOptimiser
from .barycentre import EnsembleBarycentreSurrogateOptimiser
from .baselines.evolutions import DifferentialEvolutionOptimizer
from .baselines.protes import ProtesOptimiser

__all__ = [
    'Optimiser',
    'SurrogateOptimiser',
    'BarycentreSurrogateOptimiser',
    'EnsembleBarycentreSurrogateOptimiser',
    'DifferentialEvolutionOptimizer',
    'ProtesOptimiser'
]



import numpy as np
from tqdm import tqdm
from .base import GenSurrogate

import torch
import torch.nn as nn
from torch.autograd import Variable
from torch.optim.lr_scheduler import StepLR
from torch.utils.data import TensorDataset, DataLoader

cuda = True if torch.cuda.is_available() else False
device = torch.device("cuda" if cuda else "cpu")
# Tensor = torch.cuda.FloatTensor if cuda else torch.FloatTensor


class Generator(nn.Module):
    def __init__(self, n_inputs, n_outputs, hidden_size=64):
        super(Generator, self).__init__()
        
        self.model = nn.Sequential(
            nn.Linear(n_inputs, hidden_size),
            nn.Tanh(),
            nn.Linear(hidden_size, n_outputs),
        )

    def forward(self, x_cond, x_noise):
        x = torch.cat((x_cond, x_noise), dim=1)
        x = self.model(x)
        return x
    

class Discriminator(nn.Module):
    def __init__(self, n_inputs, hidden_size=64):
        super(Discriminator, self).__init__()
        
        self.model = nn.Sequential(
            nn.Linear(n_inputs, hidden_size),
            nn.Tanh(),
            nn.Linear(hidden_size, 1),
        )

    def forward(self, x):
        x = self.model(x)
        return x

    
class WGAN_GP(GenSurrogate):
    def __init__(self, **kwargs):
        super(WGAN_GP, self).__init__()

        self.name        = 'GAN'
        self.G           = kwargs['G'] if 'G' in kwargs else None
        self.D           = kwargs['D'] if 'D' in kwargs else None
        self.batch_size  = kwargs['batch_size'] if 'batch_size' in kwargs else 16
        self.hidden_size = kwargs['hidden_size'] if 'hidden_size' in kwargs else 64
        self.n_epochs    = kwargs['n_epochs'] if 'n_epochs' in kwargs else 100
        self.n_disc      = kwargs['n_disc'] if 'n_disc' in kwargs else 5
        self.latent_dim  = kwargs['latent_dim'] if 'latent_dim' in kwargs else 10
        self.lambda_gp   = kwargs['lambda_gp'] if 'lambda_gp' in kwargs else 1
        self.G_lr        = kwargs['G_lr'] if 'G_lr' in kwargs else 1e-4
        self.D_lr        = kwargs['D_lr'] if 'D_lr' in kwargs else 1e-4
        self.scheduler   = kwargs['scheduler'] if 'scheduler' in kwargs else False
        self.verbose     = kwargs['verbose'] if 'verbose' in kwargs else 1
        self.save_loss   = kwargs['save_loss'] if 'save_loss' in kwargs else False

        if self.save_loss:
            self.G_loss_hist = []
            self.D_loss_hist = []
    

    def gradient_pen(self, cond_data, real_test_data, gen_data):

        alpha = torch.tensor(np.random.random((real_test_data.size(0), 1)), dtype=torch.float32, device=device)

        inter = (alpha * real_test_data + ((1 - alpha) * gen_data)).requires_grad_(True)
        D_inter = self.D(torch.cat((cond_data, inter), dim=1))
        fake = Variable(torch.ones(real_test_data.size(0), 1, dtype=torch.float32, device=device))

        gradients = torch.autograd.grad(
            outputs      = D_inter,
            inputs       = inter,
            grad_outputs = fake,
            create_graph = True,
            retain_graph = True,
            only_inputs  = True,
        )[0]
        
        gp = ((gradients.norm(2, dim=1) - 1) ** 2).mean()
        return gp
        
        
    def fit(self, X, y, **kwargs):

        self.verbose = kwargs['verbose'] if 'verbose' in kwargs else self.verbose

        if self.G is None:
            self.G = Generator(n_inputs=X.shape[-1] + self.latent_dim, n_outputs=y.shape[-1], hidden_size=self.hidden_size)
        if self.D is None:
            self.D = Discriminator(n_inputs=X.shape[-1] + y.shape[-1], hidden_size=self.hidden_size)
        
        if cuda:
            self.G.cuda()
            self.D.cuda()
        
        X = torch.tensor(X, dtype=torch.float32, device=device)
        y = torch.tensor(y, dtype=torch.float32, device=device)
        X_train = TensorDataset(X, y)
        
        if self.scheduler:
            self.G_opt = torch.optim.Adam(self.G.parameters(), lr=self.G_lr*1e2)
            self.D_opt = torch.optim.Adam(self.D.parameters(), lr=self.D_lr*1e2)
            
            G_scheduler = StepLR(self.G_opt, step_size=self.n_epochs//3, gamma=0.1)
            D_scheduler = StepLR(self.D_opt, step_size=self.n_epochs//3, gamma=0.1)
        else:
            self.G_opt = torch.optim.Adam(self.G.parameters(), lr=self.G_lr)
            self.D_opt = torch.optim.Adam(self.D.parameters(), lr=self.D_lr)
        
        self.G.train(True)
        self.D.train(True)
        
        fit_loop = range(self.n_epochs)
        if self.verbose > 1:
            fit_loop = tqdm(fit_loop, unit="epoch", leave=True, position=2)
        
        for e in fit_loop:
            for X_batch, y_batch in DataLoader(X_train, batch_size=self.batch_size, shuffle=True, drop_last=True):

                for _ in range(self.n_disc):
                    noise_batch = Variable(torch.tensor(np.random.normal(0, 1, (self.batch_size, self.latent_dim)), dtype=torch.float32, device=device))

                    y_gen_batch = self.G(X_batch, noise_batch)

                    D_real = self.D(torch.cat((X_batch, y_batch), dim=1))
                    D_fake = self.D(torch.cat((X_batch, y_gen_batch.detach()), dim=1))
                    
                    D_loss = torch.mean(D_fake) - torch.mean(D_real)
                    grad_p = self.gradient_pen(X_batch, y_batch, y_gen_batch.detach())
                    D_loss += self.lambda_gp * grad_p

                    self.D_opt.zero_grad()
                    D_loss.backward()
                    self.D_opt.step()
                    
                    if self.save_loss:
                        self.D_loss_hist.append(D_loss.detach().numpy())
                
                noise_batch = Variable(torch.tensor(np.random.normal(0, 1, (self.batch_size, self.latent_dim)), dtype=torch.float32, device=device))

                y_gen_batch = self.G(X_batch, noise_batch)
                
                D_fake = self.D(torch.cat((X_batch, y_gen_batch), dim=1))
                
                G_loss = -torch.mean(D_fake)
            
                self.G_opt.zero_grad()
                G_loss.backward()
                self.G_opt.step()
                
                if self.save_loss:
                    self.G_loss_hist.append(G_loss.detach().numpy())
            
            if self.verbose > 1:
                fit_loop.set_description(f"G loss: {G_loss:.4f}, D loss: {D_loss:.4f}")
            
            if (self.scheduler):
                G_scheduler.step()
                D_scheduler.step()
        
        self.scheduler = True
        self.G.train(False)
        self.D.train(False)
    
    
    def sample(self, X_cond):
        noise = Variable(torch.tensor(np.random.normal(0, 1, (X_cond.shape[0], self.latent_dim)), dtype=torch.float32, device=device))
        X_gen = self.G(torch.tensor(X_cond, dtype=torch.float32, device=device), noise)
        return X_gen.cpu().detach().numpy()
